' Copyright (c) Microsoft Corporation.  All rights reserved.

'------------------------------------------------------------------------------
' PLEASE NOTE:  Microsoft Corporation (or based on where you live, one of its affiliates) 
' licenses this supplement to you.  You may use it with each validly licensed copy of Microsoft 
' Visual Studio Pre-Release Code-Name "ORCAS" software (the “software”).  You may not use the 
' supplement if you do not have a license for the software.  The license terms for the software 
' apply to your use of this supplement. 
'
' Please note: As this software is distributed in Quebec, Canada, this notice is provided below in French.
' Remarque: Ce logiciel étant distribué au Québec, Canada, les termes de cette licence sont fournis ci-dessous en français.
' REMARQUE:  Microsoft Corporation (ou en fonction du lieu où vous vivez, l’un de ses affiliés) vous 
' accorde une licence pour ce supplément.  Vous pouvez utiliser une copie de ce supplément avec chaque 
' copie concédée sous licence du logiciel Microsoft Visual Studio Pre-Release Code-Name "ORCAS" (le « logiciel »).  
' Vous n’êtes pas autorisé à utiliser ce supplément si vous n’êtes pas titulaire d’une licence pour le logiciel.  
' Les conditions de licence pour le logiciel s’appliquent à l’utilisation de ce supplément.  
'------------------------------------------------------------------------------
Imports System.Configuration
Imports NorthwindLib
Imports System.Data.Objects
Imports System.Data.Common
Imports System.Data.EntityClient
Imports System.Collections.Generic


Module Sample

    Sub Main()

        ' Queries
        SimpleQuery()
        Projection()
        QueryWithParams()
        Inheritance_PolymorphicResultSets()
        Inheritance_FilterByType()


        ' Collections and relationships
        RelationshipTraversal()
        LazyLoadingRelatedEntities()

        ' Update & graph maintenance, concurrency checks
        UpdatesAndFixUps()
        OptimisticConcurrency()
    End Sub

    ' Execute a really simple query against the EDM schema to
    ' fetch all the Category entities from the Categories entity set
    Sub SimpleQuery()
        ' Northwind represents the ADO.NET object context that acts
        ' as a factory for queries, tracks object state, and is
        ' used to initiate updates against the database.

        ' The Northwind class is generated by the system from the
        ' .csdl file, and represents the entity container, exposing
        ' one member for each of the entity-sets (e.g. Categories, 
        ' Products, etc.)
        Using db As New Northwind()

            Dim c As Category
            ' Query for all of the Category instances and return them
            ' as .NET objects
            Dim categories As ObjectQuery(Of Category) = db.CreateQuery(Of Category)( _
                "SELECT VALUE c FROM Categories AS c ORDER BY c.CategoryName")

            ' by doing a foreach over the query object it gets executed 
            ' automatically. If a more explicit model is preferred, the Execute()
            ' method can be used.
            For Each c In categories
                Console.WriteLine("{0}\t{1}", c.CategoryName, c.Description)
            Next
        End Using

        Separator()
    End Sub

    ' When projecting arbitrary columns in a query, the result-set is a 
    ' sequence of records instead of a sequence of instances of a particular 
    ' class, because the projection doesn't match any class. Note that this
    ' is different from LINQ, were the system actually CAN return .NET objects
    ' from projections, represented as anonymous types (see the ObjectServicesLINQ
    ' for examples).
    Sub Projection()
        Using db As New Northwind()

            Dim categoryRec As DbDataRecord

            ' Project out the name and description for each of
            ' the categories. Since the result type is not a known
            ' .NET type, we use records.
            Dim categories As ObjectQuery(Of DbDataRecord) = db.CreateQuery(Of DbDataRecord)( _
                "SELECT c.CategoryName, c.Description FROM Categories AS c ORDER BY c.CategoryName")

            For Each categoryRec In categories
                ' Since the result is a set of records we have to use late binding
                ' to get the entity property values
                Console.WriteLine("{0}" & vbTab & "{1}", categoryRec("CategoryName"), categoryRec("Description"))
            Next
        End Using

        Separator()
    End Sub

    ' Use parameters in a query to include external values such as user input into a query
    Sub QueryWithParams()
        Using db = New Northwind()

            Dim c As Category
            ' Query for a particular Category for which we already
            ' know the ID. Pass the ID as a parameter to the query.
            Dim categories As ObjectQuery(Of Category) = db.CreateQuery(Of Category)( _
                "SELECT VALUE c FROM Categories AS c WHERE c.CategoryID = @id", _
                New ObjectParameter("id", 1))

            ' by doing a foreach over the query object it gets executed 
            ' automatically. If a more explicit model is preferred, the Execute()
            ' method can be used.
            For Each c In categories
                Console.WriteLine("{0}" & vbTab & "{1}", c.CategoryName, c.Description)
            Next
        End Using

        Separator()
    End Sub

    ' The Products entity-set has instances of the Product and DiscontinuedProduct
    ' entity types (DiscontinuedProduct is a subtype of Product). Formulate a query
    ' that returns instances of both and discriminate them in the client
    Sub Inheritance_PolymorphicResultSets()
        Using db = New Northwind()

            Dim p As Product

            Dim products As ObjectQuery(Of Product) = db.CreateQuery(Of Product)( _
                "SELECT VALUE p FROM Products AS p")

            For Each p In products
                ' if we're only accessing properties that are defined in the most base
                ' type (e.g. Product in this case), then no special casing is needed however,
                ' here we want to access a property that's only available for a specific
                ' subtype (the UnitsInStock property that's only in DiscontinuedProduct). So
                ' we use the .NET type information to decide which way to go at runtime.

                ' Print out properties that are present in all instances of Product or its subtypes
                Console.Write("{0}" & vbTab & "{1}", p.ProductName, p.UnitPrice)

                Dim discProd As DiscontinuedProduct = TryCast(p, DiscontinuedProduct)
                If discProd IsNot Nothing Then
                    Console.WriteLine(vbTab & "{0}", discProd.UnitsInStock)
                Else
                    Console.WriteLine()
                End If
            Next

            Separator()

            ' Alternate approach that may apply depending on the scenario:

            ' When a given action is available in all subtypes of a hierarchy, having
            ' different behavior depending on the subtype, that can be modeled as a virtual
            ' method that can be added to the partial classes. See Product.cs and DiscontinuedProduct.cs
            ' to see the implementation of StockInfo().

            ' (note that by doing another foreach over the query we're re-executing it)
            For Each p In products
                ' no type-based special case now, just call the virtual GetStockInfo function
                Console.WriteLine("{0}" & vbTab & "{1}" & vbTab & "{2}", p.GetStockInfo(), p.ProductName, p.UnitPrice)
            Next
        End Using

        Separator()
    End Sub

    ' The Products entity-set has instances of the Product and DiscontinuedProduct
    ' entity types (DiscontinuedProduct is a subtype of Product). Formulate a query
    ' that returns only instances of DiscontinuedProduct or one of its subtypes
    Sub Inheritance_FilterByType()
        Using db = New Northwind()

            Dim p As DiscontinuedProduct

            ' Note that since in the Entity SQL statement we're restricting the
            ' results to be only DiscontinuedProduct, we can use DiscontinuedProduct
            ' to also be the type of the query (Product would also work, but if you'd
            ' want to access properies that are specific to the subtypes then you'd
            ' have to use a cast).
            Dim products As ObjectQuery(Of Product) = db.CreateQuery(Of Product)( _
                "SELECT VALUE p FROM Products AS p " & _
                "WHERE p IS OF (NorthwindLib.DiscontinuedProduct)")

            For Each p In products
                Console.WriteLine("{0}" & vbTab & "{1}" & vbTab & "{2}", p.ProductName, p.UnitPrice, p.UnitsInStock)
            Next
        End Using

        Separator()
    End Sub

    ' Use the associations declared in the model to easily navigate through relationships. Here
    ' we use the "." operator to traverse the relationship between categories and
    ' products that's described in the model (the association is called "Category_Product"
    ' in the EDM model)
    Sub RelationshipTraversal()
        Using db = New Northwind()

            Dim p As Product

            ' Find all products from the "Beverages" category 
            Dim beverages As ObjectQuery(Of Product) = db.CreateQuery(Of Product)( _
                "SELECT VALUE p FROM Products AS p WHERE p.Category.CategoryName = @name", _
                New ObjectParameter("name", "Beverages"))

            For Each p In beverages
                Console.WriteLine(p.ProductName)
            Next
        End Using

        Separator()
    End Sub

    ' When fetching an entity that has other related entities, those related entities
    ' can be loaded later on as needed. For relationships with cardinality greater than
    ' 1, the entity will contain a collection representing the relationship if the
    ' cardinality is 1, then it'll contain a reference to the other entity and a Ref<T>
    ' member that's useful for loading it.
    Sub LazyLoadingRelatedEntities()
        Using db = New Northwind()

            Dim p As Product

            ' Find all products from the "Beverages" category 
            Dim beverages As ObjectQuery(Of Product) = db.CreateQuery(Of Product)( _
                "SELECT VALUE p FROM Products AS p WHERE p.Category.CategoryName = @name", _
                New ObjectParameter("name", "Beverages"))

            For Each p In beverages
                Console.WriteLine(p.ProductName)

                ' this product object doesn't have the category loaded. If we
                ' want the Category entity that's associated with this Product we can 
                ' load it:
                p.CategoryReference.Load()

                ' now that we loaded the reference, we can use the p.Category product
                Console.WriteLine(vbTab & "Category: {0}", p.Category.CategoryName)
            Next

            Separator()

            ' Going in the oposite direction
            ' Given a category, we can find the collection of products that's
            ' related to that category

            ' find a specific category (since we know this query returns 1 entity
            ' we can use the First() method):
            Dim cat As Category = db.CreateQuery(Of Category)( _
                "SELECT VALUE c FROM Categories AS c WHERE c.CategoryID = @id", _
                New ObjectParameter("id", 1)).First()

            Console.WriteLine("{0} ({1})", cat.CategoryName, cat.Description)

            ' now load the products for this category
            cat.Products.Load()

            ' if you try this foreach before calling load you would see an empty collection
            For Each p In cat.Products
                Console.WriteLine(vbTab & p.ProductName)
            Next

            Separator()

            ' Querying a related set independently of the local collection

            ' This fetches a specific subset of the products for a given category
            ' (".Source" here represents the query that returns the products related
            ' to this category instance). Also note the use of query construction method
            ' "where" to compose the query
            Dim prods As ObjectQuery(Of Product) = cat.Products.Source.Where("it.UnitPrice > 20")

            For Each p In prods
                Console.WriteLine("{0}" & vbTab & "{1}", p.ProductName, p.UnitPrice)
            Next
        End Using

        Separator()
    End Sub

    ' The Entity Framework automatically handles updates. It tracks changes made to
    ' the .NET objects that represent entities, and when SaveChanges() is called
    ' it generates the requires SQL DML statements to send the updates to the 
    ' database. Also, while changes that affect relationships are done in the client
    ' the references and collections are automatically synchronized to maintain
    ' the consistency of the object graph.
    Sub UpdatesAndFixUps()

        Dim productId As Integer

        Using db = New Northwind()

            ' find a specific category (since we know this query returns 1 entity
            ' we can use the First() method):
            Dim cat As Category = db.CreateQuery(Of Category)( _
                "SELECT VALUE c FROM Categories AS c WHERE c.CategoryID = @id", _
                New ObjectParameter("id", 1)).First()

            ' create a new product and associate it with this category
            Dim p As New Product()
            p.ProductName = "ADO.NET Branded Cola :)"
            p.UnitPrice = CType(1.99, Decimal)

            ' Set this product to be in the "Beverages" category (the one we retrieved 
            ' above). After this assignment, the Products collection in the Category
            ' object will also be fixed up, so this product will be present in the
            ' collection.
            p.Category = cat

            ' Now push the change to the database. Note that since the table is
            ' set to automatically generate keys (ProductID is an IDENTITY column),
            ' the ProductID property of "p" will be updated after the call to SaveChanges()
            db.SaveChanges()
            Console.WriteLine("New product created, ProductID: " & p.ProductID)

            ' save the product id
            productId = p.ProductID
        End Using

        ' Now find and remove the product we just created (we could have
        ' use the same context (Northwind) a new one is used here just 
        ' to introduce variations
        Using db = New Northwind()
            ' use the productId we saved before to find the product entity. This is
            ' another way of fetching an entity by key, creating an entity key instead
            ' of using an Entity SQL query that goes "where productid = ..."

            Dim keys(0) As KeyValuePair(Of String, Object)
            keys(0) = New KeyValuePair(Of String, Object)("ProductID", productId)

            Dim p As Product = CType(db.GetObjectByKey(db.CreateKey(GetType(Product), keys)), Product)

            ' now delete the new product and push the change to the database
            p.CategoryReference.Load()
            db.DeleteObject(p)
            db.SaveChanges()
        End Using

        Separator()
    End Sub

    ' As part of the update infrastructure, the Entity Framework includes
    ' very flexible optimistic concurrency support. The author of an EDM
    ' schema can individually specify which properties of an entity should
    ' be considered for conflict detection. These could be regular properties
    ' or a timestamp used specifically for this purpose. Look for the 
    ' "concurrency" attributes in the CSDL file to see how properties
    ' are marked for concurrency checks.
    Sub OptimisticConcurrency()

        Dim categoryId As Integer

        ' create a category using a separate context
        Using db As New Northwind()
            Dim cat As New Category()
            cat.CategoryName = "SampleCateg"
            cat.Description = "This is an example"
            db.AddObject(cat)
            db.SaveChanges()

            ' grab the server-generated key for this category
            categoryId = cat.CategoryID
        End Using

        ' we'll use two object contexts to simulate two
        ' concurrent sessions
        Using db1 As New Northwind(), db2 As New Northwind()

            ' find the category through the first session, and make a change to it
            Dim cat1 As Category = db1.CreateQuery(Of Category)( _
                "SELECT VALUE c FROM Categories AS c WHERE c.CategoryID = @id", _
                New ObjectParameter("id", categoryId)).First()
            cat1.CategoryName = "from db1"

            ' now find the category and modify it through the second session (db2), while
            ' the first session (db1) still has pending modifications
            Dim cat2 As Category = db2.CreateQuery(Of Category)( _
                "SELECT VALUE c FROM Categories AS c WHERE c.CategoryID = @id", _
                New ObjectParameter("id", categoryId)).First()
            cat2.CategoryName = "from db2"

            ' now save the changes we made through db2
            db2.SaveChanges()

            ' at this point, the initial values of this category that db1
            ' saw have changed, so the following update should fail
            Dim conflictCaught As Boolean = False
            Try
                db1.SaveChanges()
            Catch ex As OptimisticConcurrencyException
                conflictCaught = True
                Console.WriteLine("Update conflict successfully detected")
            End Try

            If Not conflictCaught Then Throw New Exception("Didn't catch update conflict")

            ' delete the category we created for the example
            db2.DeleteObject(cat2)
            db2.SaveChanges()
        End Using

        Separator()
    End Sub

#Region "Helper Functions"

    Sub Separator()
        Console.WriteLine("------------------------------------------------")
    End Sub

#End Region
End Module
